import * as plugins from './plugins.js';

/**
 * Optional path pattern configuration that can be added to proxy configs
 */
export interface IPathPatternConfig {
  pathPattern?: string;
}

/**
 * Interface for router result with additional metadata
 */
export interface IRouterResult {
  config: plugins.tsclass.network.IReverseProxyConfig;
  pathMatch?: string;
  pathParams?: Record<string, string>;
  pathRemainder?: string;
}

export class ProxyRouter {
  // Using a Map for O(1) hostname lookups instead of array search
  private hostMap: Map<string, plugins.tsclass.network.IReverseProxyConfig[]> = new Map();
  // Store original configs for reference
  private reverseProxyConfigs: plugins.tsclass.network.IReverseProxyConfig[] = [];
  // Default config to use when no match is found (optional)
  private defaultConfig?: plugins.tsclass.network.IReverseProxyConfig;
  // Store path patterns separately since they're not in the original interface
  private pathPatterns: Map<plugins.tsclass.network.IReverseProxyConfig, string> = new Map();

  constructor(
    configs?: plugins.tsclass.network.IReverseProxyConfig[],
    private readonly logger: { 
      error: (message: string, data?: any) => void;
      warn: (message: string, data?: any) => void;
      info: (message: string, data?: any) => void;
      debug: (message: string, data?: any) => void;
    } = console
  ) {
    if (configs) {
      this.setNewProxyConfigs(configs);
    }
  }

  /**
   * Sets a new set of reverse configs to be routed to
   * @param reverseCandidatesArg Array of reverse proxy configurations
   */
  public setNewProxyConfigs(reverseCandidatesArg: plugins.tsclass.network.IReverseProxyConfig[]): void {
    this.reverseProxyConfigs = [...reverseCandidatesArg];
    
    // Reset the host map and path patterns
    this.hostMap.clear();
    this.pathPatterns.clear();
    
    // Find default config if any (config with "*" as hostname)
    this.defaultConfig = this.reverseProxyConfigs.find(config => config.hostName === '*');
    
    // Group configs by hostname for faster lookups
    for (const config of this.reverseProxyConfigs) {
      // Skip the default config as it's stored separately
      if (config.hostName === '*') continue;
      
      const hostname = config.hostName.toLowerCase(); // Case-insensitive hostname lookup
      
      if (!this.hostMap.has(hostname)) {
        this.hostMap.set(hostname, []);
      }
      
      // Check for path pattern in extended properties 
      // (using any to access custom properties not in the interface)
      const extendedConfig = config as any;
      if (extendedConfig.pathPattern) {
        this.pathPatterns.set(config, extendedConfig.pathPattern);
      }
      
      // Add to the list of configs for this hostname
      this.hostMap.get(hostname).push(config);
    }
    
    // Sort configs for each hostname by specificity
    // More specific path patterns should be checked first
    for (const [hostname, configs] of this.hostMap.entries()) {
      if (configs.length > 1) {
        // Sort by pathPattern - most specific first 
        // (null comes last, exact paths before patterns with wildcards)
        configs.sort((a, b) => {
          const aPattern = this.pathPatterns.get(a);
          const bPattern = this.pathPatterns.get(b);
          
          // If one has a path and the other doesn't, the one with a path comes first
          if (!aPattern && bPattern) return 1;
          if (aPattern && !bPattern) return -1;
          if (!aPattern && !bPattern) return 0;
          
          // Both have path patterns - more specific (longer) first
          // This is a simple heuristic; we could use a more sophisticated approach
          return bPattern.length - aPattern.length;
        });
      }
    }
    
    this.logger.info(`Router initialized with ${this.reverseProxyConfigs.length} configs (${this.hostMap.size} unique hosts)`);
  }

  /**
   * Routes a request based on hostname and path
   * @param req The incoming HTTP request
   * @returns The matching proxy config or undefined if no match found
   */
  public routeReq(req: plugins.http.IncomingMessage): plugins.tsclass.network.IReverseProxyConfig {
    const result = this.routeReqWithDetails(req);
    return result ? result.config : undefined;
  }
  
  /**
   * Routes a request with detailed matching information
   * @param req The incoming HTTP request
   * @returns Detailed routing result including matched config and path information
   */
  public routeReqWithDetails(req: plugins.http.IncomingMessage): IRouterResult | undefined {
    // Extract and validate host header
    const originalHost = req.headers.host;
    if (!originalHost) {
      this.logger.error('No host header found in request');
      return this.defaultConfig ? { config: this.defaultConfig } : undefined;
    }
    
    // Parse URL for path matching
    const urlPath = new URL(
      req.url || '/', 
      `http://${originalHost}`
    ).pathname;
    
    // Extract hostname without port
    const hostWithoutPort = originalHost.split(':')[0].toLowerCase();
    
    // Find configs for this hostname
    const configs = this.hostMap.get(hostWithoutPort);
    
    if (configs && configs.length > 0) {
      // Check each config for path matching
      for (const config of configs) {
        // Get the path pattern if any
        const pathPattern = this.pathPatterns.get(config);
        
        // If no path pattern specified, this config matches all paths
        if (!pathPattern) {
          return { config };
        }
        
        // Check if path matches the pattern
        const pathMatch = this.matchPath(urlPath, pathPattern);
        if (pathMatch) {
          return {
            config,
            pathMatch: pathMatch.matched,
            pathParams: pathMatch.params,
            pathRemainder: pathMatch.remainder
          };
        }
      }
    }
    
    // Try wildcard subdomains if no direct match found
    // For example, if request is for sub.example.com, try *.example.com
    const domainParts = hostWithoutPort.split('.');
    if (domainParts.length > 2) {
      const wildcardDomain = `*.${domainParts.slice(1).join('.')}`;
      const wildcardConfigs = this.hostMap.get(wildcardDomain);
      
      if (wildcardConfigs && wildcardConfigs.length > 0) {
        // Use the first matching wildcard config
        // Could add path matching logic here as well
        return { config: wildcardConfigs[0] };
      }
    }
    
    // Fall back to default config if available
    if (this.defaultConfig) {
      this.logger.warn(`No specific config found for host: ${hostWithoutPort}, using default`);
      return { config: this.defaultConfig };
    }
    
    this.logger.error(`No config found for host: ${hostWithoutPort}`);
    return undefined;
  }
  
  /**
   * Sets a path pattern for an existing config
   * @param config The existing configuration
   * @param pathPattern The path pattern to set
   * @returns Boolean indicating if the config was found and updated
   */
  public setPathPattern(
    config: plugins.tsclass.network.IReverseProxyConfig, 
    pathPattern: string
  ): boolean {
    const exists = this.reverseProxyConfigs.includes(config);
    if (exists) {
      this.pathPatterns.set(config, pathPattern);
      return true;
    }
    return false;
  }

  /**
   * Matches a URL path against a pattern
   * Supports:
   * - Exact matches: /users/profile
   * - Wildcards: /api/* (matches any path starting with /api/)
   * - Path parameters: /users/:id (captures id as a parameter)
   * 
   * @param path The URL path to match
   * @param pattern The pattern to match against
   * @returns Match result with params and remainder, or null if no match
   */
  private matchPath(path: string, pattern: string): { 
    matched: string; 
    params: Record<string, string>; 
    remainder: string;
  } | null {
    // Handle exact match
    if (path === pattern) {
      return {
        matched: pattern,
        params: {},
        remainder: ''
      };
    }
    
    // Handle wildcard match
    if (pattern.endsWith('/*')) {
      const prefix = pattern.slice(0, -2);
      if (path === prefix || path.startsWith(`${prefix}/`)) {
        return {
          matched: prefix,
          params: {},
          remainder: path.slice(prefix.length)
        };
      }
      return null;
    }
    
    // Handle path parameters
    const patternParts = pattern.split('/');
    const pathParts = path.split('/');
    
    // Check if paths are compatible length
    if (
      // If pattern doesn't end with wildcard, paths must have the same number of parts
      (!pattern.endsWith('/*') && patternParts.length !== pathParts.length) ||
      // If pattern ends with wildcard, path must have at least as many parts as the pattern
      (pattern.endsWith('/*') && pathParts.length < patternParts.length - 1)
    ) {
      return null;
    }
    
    const params: Record<string, string> = {};
    const matchedParts: string[] = [];
    
    // Compare path parts
    for (let i = 0; i < patternParts.length; i++) {
      const patternPart = patternParts[i];
      
      // Handle wildcard at the end
      if (patternPart === '*' && i === patternParts.length - 1) {
        break;
      }
      
      // If pathParts[i] doesn't exist, we've reached the end of the path
      if (i >= pathParts.length) {
        return null;
      }
      
      const pathPart = pathParts[i];
      
      // Handle parameter
      if (patternPart.startsWith(':')) {
        const paramName = patternPart.slice(1);
        params[paramName] = pathPart;
        matchedParts.push(pathPart);
        continue;
      }
      
      // Handle exact match for this part
      if (patternPart !== pathPart) {
        return null;
      }
      
      matchedParts.push(pathPart);
    }
    
    // Calculate the remainder
    let remainder = '';
    if (pattern.endsWith('/*')) {
      remainder = '/' + pathParts.slice(patternParts.length - 1).join('/');
    }
    
    return {
      matched: matchedParts.join('/'),
      params,
      remainder
    };
  }
  
  /**
   * Gets all currently active proxy configurations
   * @returns Array of all active configurations
   */
  public getProxyConfigs(): plugins.tsclass.network.IReverseProxyConfig[] {
    return [...this.reverseProxyConfigs];
  }
  
  /**
   * Gets all hostnames that this router is configured to handle
   * @returns Array of hostnames
   */
  public getHostnames(): string[] {
    return Array.from(this.hostMap.keys());
  }
  
  /**
   * Adds a single new proxy configuration
   * @param config The configuration to add
   * @param pathPattern Optional path pattern for route matching
   */
  public addProxyConfig(
    config: plugins.tsclass.network.IReverseProxyConfig, 
    pathPattern?: string
  ): void {
    this.reverseProxyConfigs.push(config);
    
    // Store path pattern if provided
    if (pathPattern) {
      this.pathPatterns.set(config, pathPattern);
    }
    
    this.setNewProxyConfigs(this.reverseProxyConfigs);
  }
  
  /**
   * Removes a proxy configuration by hostname
   * @param hostname The hostname to remove
   * @returns Boolean indicating whether any configs were removed
   */
  public removeProxyConfig(hostname: string): boolean {
    const initialCount = this.reverseProxyConfigs.length;
    this.reverseProxyConfigs = this.reverseProxyConfigs.filter(
      config => config.hostName !== hostname
    );
    
    if (initialCount !== this.reverseProxyConfigs.length) {
      this.setNewProxyConfigs(this.reverseProxyConfigs);
      return true;
    }
    
    return false;
  }
}